% $Id: Memo.lhs,v 1.8 2004/10/06 11:16:40 ross Exp $
%
% (c) The GHC Team, 1999
%
% Hashing memo tables.


This module originally implemented lazy memo functions.

We have replaced the pointer equality test that was performed in the
original module by a structural equality test.

In this way, we obtained a memo primitive that we can use in the
context of our AG embedding.

\begin{code}
{-# OPTIONS -fglasgow-exts #-}

module Memo

	(memo  	-- :: (a -> b) -> a -> b
	) 

	where


import System.Mem.Weak		( Weak, mkWeakPtr, mkWeak, deRefWeak, finalize )
import Data.Array.IO		( IOArray, newArray, readArray, writeArray )
import System.IO.Unsafe		( unsafePerformIO )
import Control.Concurrent.MVar	( MVar, newMVar, putMVar, takeMVar )

import qualified Data.Map as Map

import Debug.Trace

\end{code}

-----------------------------------------------------------------------------
Memo table representation.

...

\begin{code}

type MemoTable key val
	= MVar (Map.Map key (Weak val))


\end{code}

We use an MVar to the hash table, so that several threads may safely
access it concurrently.  This includes the finalization threads that
remove entries from the table.

ToDo: Can efficiency be improved at all?

\begin{code}

-- memo :: (Ord a) => (a -> b) -> a -> b
memo f = strict (lazymemo f)

strict = ($!)

-- lazymemo :: Ord k => (k -> val) -> k -> val
lazymemo f
	= let (table, weak)
			= unsafePerformIO (
				do {tbl <- return Map.empty;
				    mvar <- newMVar tbl;
				    weak <- mkWeakPtr mvar (Just (table_finalizer tbl));
				    return (mvar, weak)
				    })
	  in memo' f table weak


table_finalizer :: (Map.Map key (Weak val)) -> IO ()
table_finalizer table
	= do {-trace "I'm cleaning a table!!\n\n\n" $-}
	    sequence_ [finalize w | w <- Map.elems table]


-- memo' :: (Ord a) => (a -> b) -> MemoTable a b 
--			     -> Weak (MemoTable a b) -> a -> b
memo' f ref weak_ref
	= \k -> unsafePerformIO $
			do table <- takeMVar ref
			   lkp   <- lookupK k table
			   case lkp of 
                                     Just result 
					-> do putMVar ref table
					      return result
				     Nothing
					-> do let result = f k
					      weak <- mkWeak k result (Just (finalizer k weak_ref))
					      table' <- return $ Map.insert k weak table
					      putMVar ref table'
					      return result


finalizer :: Ord k => k -> Weak (MVar (Map.Map k a)) -> IO ()
finalizer k weak_ref
	= do r <- deRefWeak weak_ref
	     case r of 
		Nothing 
			-> return ()
		Just mvar
			-> do table 
				<- takeMVar mvar
			      table' 
				<- {- trace "I'm cleaning an entry!!" $-}
				   return (Map.delete k table)
			      putMVar mvar table'


-- lookupK :: Ord k => k -> Map.Map k (Weak v) -> IO (Maybe v)
lookupK k fmap
	= do v <- {- trace ("searching for " ++(show k) ++ "in " ++ (concatMap show (Map.keys fmap))) -} 
                  return (Map.lookup k fmap)
	     case v of
		Nothing 
			-> trace "cache miss!!\n" $ return Nothing
		Just weak
			-> do maybe_item <- deRefWeak weak
			      case maybe_item of
					Nothing 
						-> error "dead weak pair"
					Just val 
						-> trace "chache hit!!\n" $ return (Just val)


\end{code}
